package com.sun.org.apache.bcel.internal.classfile;

import java.util.HashMap;
import com.sun.org.apache.bcel.internal.Const;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.DataOutputStream;
import java.util.Map;

public abstract class Attribute implements Cloneable, Node
{
    private int name_index;
    private int length;
    private final byte tag;
    private ConstantPool constant_pool;
    private static final Map<String, Object> readers;
    
    protected Attribute(byte tag, int name_index, int length, ConstantPool constant_pool) {
        this.tag = tag;
        this.name_index = name_index;
        this.length = length;
        this.constant_pool = constant_pool;
    }
    
    @Override
    public abstract void accept(Visitor p0);
    
    public void dump(DataOutputStream file) throws IOException {
        file.writeShort(this.name_index);
        file.writeInt(this.length);
    }
    
    public static void addAttributeReader(String name, UnknownAttributeReader r) {
        Attribute.readers.put(name, r);
    }
    
    public static void removeAttributeReader(String name) {
        Attribute.readers.remove(name);
    }
    
    public static Attribute readAttribute(DataInputStream file, ConstantPool constant_pool) throws IOException, ClassFormatException {
        return readAttribute(file, constant_pool);
    }
    
    public static Attribute readAttribute(DataInput file, ConstantPool constant_pool) throws IOException, ClassFormatException {
        byte tag = -1;
        int name_index = file.readUnsignedShort();
        ConstantUtf8 c = (ConstantUtf8)constant_pool.getConstant(name_index, (byte)1);
        String name = c.getBytes();
        int length = file.readInt();
        byte i = 0;
        while (i < 22) {
            if (name.equals(Const.getAttributeName((int)i))) {
                tag = i;
                break;
            }
            ++i;
        }
        switch (tag) {
            case -1: {
                Object r = Attribute.readers.get(name);
                if ((r instanceof UnknownAttributeReader)) {
                    return ((UnknownAttributeReader)r).createAttribute(name_index, length, file, constant_pool);
                }
                return new Unknown(name_index, length, file, constant_pool);
            }
            case 1: {
                return new ConstantValue(name_index, length, file, constant_pool);
            }
            case 0: {
                return new SourceFile(name_index, length, file, constant_pool);
            }
            case 2: {
                return new Code(name_index, length, file, constant_pool);
            }
            case 3: {
                return new ExceptionTable(name_index, length, file, constant_pool);
            }
            case 4: {
                return new LineNumberTable(name_index, length, file, constant_pool);
            }
            case 5: {
                return new LocalVariableTable(name_index, length, file, constant_pool);
            }
            case 6: {
                return new InnerClasses(name_index, length, file, constant_pool);
            }
            case 7: {
                return new Synthetic(name_index, length, file, constant_pool);
            }
            case 8: {
                return new Deprecated(name_index, length, file, constant_pool);
            }
            case 9: {
                return new PMGClass(name_index, length, file, constant_pool);
            }
            case 10: {
                return new Signature(name_index, length, file, constant_pool);
            }
            case 11: {
                System.err.println("Warning: Obsolete StackMap attribute ignored.");
                return new Unknown(name_index, length, file, constant_pool);
            }
            case 12: {
                return new RuntimeVisibleAnnotations(name_index, length, file, constant_pool);
            }
            case 13: {
                return new RuntimeInvisibleAnnotations(name_index, length, file, constant_pool);
            }
            case 14: {
                return new RuntimeVisibleParameterAnnotations(name_index, length, file, constant_pool);
            }
            case 15: {
                return new RuntimeInvisibleParameterAnnotations(name_index, length, file, constant_pool);
            }
            case 16: {
                return new AnnotationDefault(name_index, length, file, constant_pool);
            }
            case 17: {
                return new LocalVariableTypeTable(name_index, length, file, constant_pool);
            }
            case 18: {
                return new EnclosingMethod(name_index, length, file, constant_pool);
            }
            case 19: {
                return new StackMap(name_index, length, file, constant_pool);
            }
            case 20: {
                return new BootstrapMethods(name_index, length, file, constant_pool);
            }
            case 21: {
                return new MethodParameters(name_index, length, file, constant_pool);
            }
            default: {
                throw new IllegalStateException("Unrecognized attribute type tag parsed: " + (int)tag);
            }
        }
    }
    
    public String getName() {
        ConstantUtf8 c = (ConstantUtf8)this.constant_pool.getConstant(this.name_index, (byte)1);
        return c.getBytes();
    }
    
    public final int getLength() {
        return this.length;
    }
    
    public final void setLength(int length) {
        this.length = length;
    }
    
    public final void setNameIndex(int name_index) {
        this.name_index = name_index;
    }
    
    public final int getNameIndex() {
        return this.name_index;
    }
    
    public final byte getTag() {
        return this.tag;
    }
    
    public final ConstantPool getConstantPool() {
        return this.constant_pool;
    }
    
    public final void setConstantPool(ConstantPool constant_pool) {
        this.constant_pool = constant_pool;
    }
    
    public Object clone() {
        Attribute attr = null;
        try {
            attr = (Attribute)super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new Error("Clone Not Supported");
        }
        return attr;
    }
    
    public abstract Attribute copy(ConstantPool p0);
    
    @Override
    public String toString() {
        return Const.getAttributeName((int)this.tag);
    }
    
    static {
        readers = new HashMap();
    }
}
